from typing import List, Dict, Tuple, Callable, Any, Optional

from drepr import DRepr, outputs
from drepr.engine import complete_description


def test_complete_description(d_s04: DRepr):
    model2subjects = {
        "d_s04": {
            ("mint:Variable:1", "dnode:Rainf_f_tavg"),
            ("mint:Variable:2", "dnode:Albedo_inst")
        }
    }

    for ds_model in [d_s04]:
        plan = complete_description(ds_model)

        # check all subjects are inferred correctly (which shows that all inferred alignments are correct)
        subjects = model2subjects[ds_model.__name__]
        assert len(subjects) > 0

        for edge in plan.sm.edges.values():
            pair = (edge.source_id, edge.target_id)
            if pair in subjects:
                assert edge.is_subject is True
                subjects.remove(pair)
        assert len(subjects) == 0



from typing import *

from drepr.engine import complete_description
from drepr.executors.cf_convention_map.cf_convention_map import CFConventionNDArrayMap
from drepr.models import Alignment, DRepr, SemanticModel, defaultdict
from drepr.outputs.array_backend.array_attr import Attribute
from drepr.outputs.array_backend.array_class import ArrayClass
from drepr.outputs.array_backend.indexed_sm import IndexedSM
from drepr.outputs.array_backend.lst_array_class import LstArrayClass
from drepr.outputs.record_id import ArrayRecordID
from drepr.outputs.base_output_sm import BaseOutputSM, identity
from drepr.outputs.base_record import BaseRecord
from drepr.outputs.base_lst_output_class import BaseLstOutputClass
from drepr.outputs.base_output_class import BaseOutputClass


class ArrayBackend(BaseOutputSM):
    """
    Array-based output. In particular, each property of a class is an array.
    A class will contains the subject property (similar to primary key column). It also has alignments between
    two subject properties of two classes and between the subject and data properties of a class, provided as
    a function alignments(source_id, target_id).

    The backend need to support a function that get a record based on some key. The key here is different to the 
    URI of a record, and it will be the index in the subject's array.
    To iter through all records of a class, we only need to loop through each index. 
    """

    def __init__(self, sm: SemanticModel, attrs: Dict[str, Attribute],
                 alignments: Dict[Tuple[str, str], List[Alignment]], inject_class_id: Callable[[str], str]):
        """
        @param sm
        @param attrs
        @param alignments get alignment from (source_id, target_id) where source_id and target_id are IDs of attributes.
        """
        # semantic model
        self.sm = sm
        # a mapping from attribute id to the attribute
        self.attrs = attrs
        # a mapping from a pair of two attributes to their alignments
        self.alignments = alignments
        # a mapping from node id to class node
        self.classes: Dict[str, ArrayClass] = {}
        self.uri2classes: Dict[str, List[ArrayClass]] = defaultdict(list)

        for c in sm.iter_class_nodes():
            self.classes[c.node_id] = ArrayClass(self, inject_class_id(c.node_id))

        for c in self.classes.values():
            c._init_schema()

        for c in self.classes.values():
            c._init_data()
            self.uri2classes[c.uri].append(c)

    @classmethod
    def from_drepr(cls, ds_model: Union[DRepr, str], resources: Union[str, Dict[str, str]], inject_class_id: Callable[[str], str]=None) -> BaseOutputSM:
        if type(ds_model) is str:
            ds_model = DRepr.parse_from_file(ds_model)

        resource_file = next(iter(resources.values())) if isinstance(resources, dict) else resources
        plan = complete_description(ds_model)
        result, attrs = CFConventionNDArrayMap.execute(ds_model, resource_file)
        return cls(plan.sm, attrs, plan.alignments, inject_class_id or identity)

    def iter_classes(self) -> Iterable[BaseOutputClass]:
        return iter(self.classes.values())

    def get_record_by_id(self, rid: ArrayRecordID) -> BaseRecord:
        return self.classes[rid.class_id].get_record_by_id(rid)

    def c(self, class_uri: str) -> BaseLstOutputClass:
        return LstArrayClass(self.uri2classes[class_uri])

    def cid(self, class_id: str) -> ArrayClass:
        return self.classes[class_id]

    def get_sm(self) -> SemanticModel:
        return self.sm


from typing import List, Dict, Tuple, Callable, Any, Optional

from drepr.outputs.array_backend.array_backend import ArrayBackend
from drepr.outputs.base_output_sm import BaseOutputSM
from drepr.outputs.record_id import GraphRecordID, BlankRecordID


def test_group_by_static_property(s01: List[BaseOutputSM], s02: List[BaseOutputSM]):
    for sm in (item for lst in [s01, s02] for item in lst):
        if isinstance(sm, ArrayBackend):
            raster_id = BlankRecordID((), sm.c("mint-geo:Raster")[0].id)
        else:
            raster_id = GraphRecordID("_:Raster1", sm.c("mint-geo:Raster")[0].id)

        gold_groups = [
            (raster_id, {
                "size": 25,
                "samples": {
                    6: {
                        "mint-geo:lat": [20.2],
                        "mint-geo:long": [44.2],
                        "rdf:value": [-0.43058764]
                    }
                }
            })
        ]
        for k, group in sm.c("mint:Variable")[0].group_by("mint:raster"):
            gk, g = gold_groups.pop(0)
            assert k == gk
            group = list(group.iter_records())
            assert len(group) == g['size']
            for i, record in g['samples'].items():
                assert group[i].m("mint-geo:lat") == record['mint-geo:lat']
                assert group[i].m("mint-geo:long") == record['mint-geo:long']
                assert group[i].m("rdf:value") == record['rdf:value']



from dataclasses import dataclass, asdict
from enum import Enum
from typing import List, Optional, Union, Set


@dataclass
class Expr:
    expr: str


@dataclass
class RangeExpr:
    start: Union[int, Expr]
    end: Optional[Union[int, Expr]]
    step: Union[int, Expr]

    def is_select_all(self) -> bool:
        return self.start == 0 and self.end is None and self.step == 1


@dataclass
class IndexExpr:
    val: Union[str, int, Expr]


@dataclass
class SetIndexExpr:
    vals: Set[Union[str, int, Expr]]


class WildcardExpr(Enum):
    Values = "*"
    Names = "*~"


StepExpr = Union[RangeExpr, IndexExpr, SetIndexExpr, WildcardExpr]


@dataclass
class Path:
    steps: List[StepExpr]

    @staticmethod
    def deserialize(raw: dict) -> "Path":
        """
        Deserialize a dictionary to get back the Path object
        :param raw:
        :return:
        """
        steps = []
        for step in raw['steps']:
            if not isinstance(step, dict):
                steps.append(WildcardExpr(step))
            elif 'start' in step:
                # range index
                start = Expr(step['start']['expr']) if isinstance(step['start'], dict) else step['start']
                end = Expr(step['end']['expr']) if isinstance(step['end'], dict) else step['end']
                step1 = Expr(step['step']['expr']) if isinstance(step['step'], dict) else step['step']
                steps.append(RangeExpr(start, end, step1))
            elif 'val' in step:
                steps.append(IndexExpr(Expr(step['val']['expr']) if isinstance(step['val'], dict) else step['val']))
            elif 'vals' in step:
                steps.append(SetIndexExpr({
                    Expr(val['expr']) if isinstance(val, dict) else val
                    for val in step['vals']
                }))
        return Path(steps)

    def to_engine_format(self) -> dict:
        steps = []
        for step in self.steps:
            if isinstance(step, RangeExpr):
                if step.end is None:
                    end = None
                elif isinstance(step.end, Expr):
                    end = f"${{{step.end.expr}}}"
                else:
                    end = step.end

                steps.append({
                    "type": "range",
                    "start": f"${{{step.start.expr}}}" if isinstance(step.start, Expr) else step.start,
                    "end": end,
                    "step": f"${{{step.step.expr}}}" if isinstance(step.step, Expr) else step.step
                })
            elif isinstance(step, IndexExpr):
                steps.append({
                    "type": "index",
                    "val": {
                        "t": "str" if isinstance(step.val, str) else "idx",
                        "c": step.val
                    }
                })
            elif isinstance(step, SetIndexExpr):
                steps.append({
                    "type": "set_index",
                    "values": [
                        {
                            "t": "str" if isinstance(val, str) else "idx",
                            "c": val
                        }
                        for val in step.vals
                    ]
                })
            elif isinstance(step, WildcardExpr):
                if step == WildcardExpr.Values:
                    steps.append({
                        "type": "wildcard"
                    })
                else:
                    raise NotImplementedError("We haven't supported operator `*~` yet")
        return {
            "steps": steps
        }

    def to_lang_format(self, use_json_path: bool = False) -> Union[list, str]:
        """
        Convert this Path object into the path object in the D-REPR language

        :param use_json_path: whether we should use the JSONPath or our new notation
        :return:
        """
        if use_json_path:
            jpath = ["$"]
            for step in self.steps:
                if isinstance(step, RangeExpr):
                    if any(isinstance(v, Expr) for v in [step.start, step.end, step.step]):
                        raise NotImplementedError("Haven't supported JSONPath with expression yet")
                    jpath.append(f"[{step.start}:{step.end or ''}:{step.step}]")
                elif isinstance(step, IndexExpr):
                    if isinstance(step.val, str):
                        jpath.append(f"['{step.val}']")
                    else:
                        jpath.append(f"[{step.val}]")
                elif isinstance(step, SetIndexExpr):
                    raise NotImplementedError()
                else:
                    jpath.append(f".{step.value}")
            return "".join(jpath)

        path = []
        for step in self.steps:
            if isinstance(step, RangeExpr):
                start, end, step = [
                    "" if v is None else (
                        f"${{{v.expr}}}" if isinstance(v, Expr) else v
                    )
                    for v in [step.start, step.end, step.step]
                ]
                path.append(f"{start}..{end}:{step}")
            elif isinstance(step, IndexExpr):
                path.append(step.val)
            elif isinstance(step, SetIndexExpr):
                path.append(step.vals)
            elif isinstance(step, WildcardExpr):
                path.append(step.value)
            else:
                raise NotImplementedError()
        return path


from collections import defaultdict
from typing import Union, Dict, List, Iterable, Callable

from drepr.engine import execute, MemoryOutput, OutputFormat
from drepr.models import DRepr, SemanticModel
from drepr.outputs.base_lst_output_class import BaseLstOutputClass
from drepr.outputs.base_output_class import BaseOutputClass
from drepr.outputs.base_output_sm import BaseOutputSM, identity
from drepr.outputs.base_record import BaseRecord
from drepr.outputs.graph_backend.graph_class import GraphClass
from drepr.outputs.graph_backend.lst_graph_class import LstGraphClass
from drepr.outputs.namespace import Namespace
from drepr.outputs.record_id import GraphRecordID


class GraphBackend(BaseOutputSM):

    def __init__(self, class2nodes: Dict[str, List[dict]], drepr: DRepr, inject_class_id: Callable[[str], str]=None):
        self.drepr = drepr
        self.sm = drepr.sm
        self.classes: Dict[str, GraphClass] = {}
        self.uri2classes: Dict[str, List[GraphClass]] = defaultdict(list)

        for class_id, nodes in class2nodes.items():
            class_id = inject_class_id(class_id)
            for u in nodes:
                u['@id'] = GraphRecordID(u['@id'], class_id)

        for c in self.sm.iter_class_nodes():
            self.classes[c.node_id] = GraphClass(self, c, class2nodes[c.node_id])
            self.uri2classes[c.label].append(self.classes[c.node_id])

    @classmethod
    def from_drepr(cls, ds_model: Union[DRepr, str], resources: Union[str, Dict[str, str]], inject_class_id: Callable[[str], str]=None) -> "GraphBackend":
        if type(ds_model) is str:
            ds_model = DRepr.parse_from_file(ds_model)

        class2nodes = execute(ds_model, resources, MemoryOutput(OutputFormat.GraphPy))
        return cls(class2nodes, ds_model, inject_class_id or identity)

    def iter_classes(self) -> Iterable[BaseOutputClass]:
        return iter(self.classes.values())

    def get_record_by_id(self, rid: GraphRecordID) -> BaseRecord:
        return self.classes[rid.class_id].get_record_by_id(rid)

    def c(self, class_uri: str) -> BaseLstOutputClass:
        return LstGraphClass(self.uri2classes[class_uri])

    def cid(self, class_id: str) -> GraphClass:
        return self.classes[class_id]

    def get_sm(self) -> SemanticModel:
        return self.sm




from pathlib import Path
from typing import List, Dict, Tuple, Callable, Any, Optional

import pytest

from drepr import DRepr


def get_drepr(dataset_dir: Path):
    return DRepr.parse_from_file(str(dataset_dir / "model.yml"))


@pytest.fixture()
def d_s04(resource_dir):
    o = get_drepr(resource_dir / "s04_shorten_gldas")
    o.__name__ = "d_s04"
    return o


@pytest.fixture()
def d_s05(resource_dir):
    o = get_drepr(resource_dir / "s05_full_gldas")
    o.__name__ = "d_s05"
    return o




import copy
from typing import Dict

import ujson
import uuid
import xmltodict

from drepr.models import DRepr, ResourceType
from drepr.patches import ResourceDataFile, ResourceData, ResourceDataString


def patch(repr: DRepr, resources: Dict[str, ResourceData]) -> DRepr:
    """
    This patch will turn any XML resources to JSON using xmltodict
    """
    need_patch = False
    for resource in repr.resources:
        if resource.type == ResourceType.XML:
            need_patch = True

    if need_patch:
        repr = copy.deepcopy(repr)
        for resource in repr.resources:
            if resource.type == ResourceType.XML:
                if isinstance(resources[resource.id], ResourceDataFile):
                    with open(resources[resource.id].file_path, "r") as f:
                        doc = xmltodict.parse(f.read())
                else:
                    doc = xmltodict.parse(resources[resource.id].value)

                resource.type = ResourceType.JSON
                resources[resource.id] = ResourceDataString(ujson.dumps(doc))
                # resources[resource.id] = ResourceDataFile(f"/tmp/{str(uuid.uuid4())}.json")
                # with open(resources[resource.id].file_path, "w") as f:
                #     ujson.dump(doc, f)

    return repr


from pathlib import Path
from typing import List, Union

import numpy as np

from PIL import Image
from PIL.TiffTags import TAGS

from drepr.executors.readers.ra_reader import NDArrayReader, Index
from drepr.models import IndexExpr, RangeExpr


class GeoTIFFReader(NDArrayReader):
    """
    Reader for GeoTIFF Resources
    """
    def __init__(self, metadata: dict, data: np.ndarray):
        self.metadata = metadata
        self.data = data

    @classmethod
    def from_file(cls, infile: str):
        with Image.open(infile) as img:
            metadata = {TAGS[key]: img.tag[key] for key in img.tag.keys()}

            # set other metadata
            metadata["filename"] = Path(infile).name
            data = np.asarray(data)
        return GeoTIFFReader(metadata, data)
    
    def len(self) -> int:
        return self.data.shape[0] + 1

    def len_range(self) -> int:
        return self.data.shape[0]

    def select(self, steps: List[Union[IndexExpr, RangeExpr]]):
        if isinstance(steps[0], IndexExpr) and steps[0].val == '@':
            # select metadata
            value = self.metadata
            for i in range(1, len(steps)):
                assert isinstance(steps[i], IndexExpr)
                value = value[steps[i].val]
            return value
        
        # select data from ndarray
        value = self.data
        for step in steps:
            if isinstance(step, IndexExpr):
                value = value[step[i].val]
            elif isinstance(step, RangeExpr):
                value = value[steps[i].start:steps[i].end:steps[i].step]
            else:
                raise NotImplementedError()
        return value
    
    def set_value(self, index: List[Index], value):
        if index[0] == '@':
            ptr = self.metadata
        else:
            ptr = self.data

        for i in range(1, len(index) - 1):
            ptr = ptr[index[i]]
        ptr[index[-1]] = value

    def get_value(self, index: List[Index]):
        if index[0] == '@':
            # read metadata
            ptr = self.metadata
        else:
            ptr = self.data

        for idx in index:
            ptr = ptr[idx]
        return ptr


from pathlib import Path
from typing import List, Union

import numpy as np
from netCDF4 import Dataset

import drepr
from drepr.executors.readers.ra_reader import NDArrayReader, Index
from drepr.models import IndexExpr, RangeExpr


class NetCDF4Reader(NDArrayReader):
    """
        Reader for NetCDF4Resource
    """

    def __init__(self, filename: str, dataset: Dataset):
        self.dataset = dataset
        # read metadata
        self.metadata = {}
        for attr in dataset.ncattrs():
            self.metadata[attr] = getattr(self.dataset, attr)
        self.metadata["filename"] = filename
        self.variables = {
            k: {
                "data": dataset.variables[k],
                "@": {
                    attr: getattr(dataset.variables[k], attr)
                    for attr in dataset.variables[k].ncattrs()
                }
            }
            for k in dataset.variables.keys()
        }

    @classmethod
    def from_file(cls, infile: str):
        ds = Dataset(infile)
        return cls(Path(infile).name, ds)

    def len(self) -> int:
        # count the global metadata at '@'
        return len(self.dataset.variables.keys()) + 1

    def len_range(self) -> int:
        raise Exception("Error in your drepr model")

    def get_value(self, index: List[Index]):
        if index[0] == '@':
            # read metadata
            ptr = self.metadata
            for i in range(1, len(index)):
                ptr = ptr[index[i]]
            return ptr

        # read variables, must be [<variable_id>, data|@, ...]
        ptr = self.variables
        for idx in index:
            ptr = ptr[idx]
        return ptr

    def set_value(self, index: List[Index], value):
        # steps.0 must be index because there is no structure change
        if index[0] == '@':
            # update metadata
            ptr = self.metadata
            for i in range(1, len(index) - 1):
                ptr = ptr[index[i]]
            ptr[index[-1]] = value
            return

        # update variables, but can only update the metadata
        assert index[1] == '@'
        ptr = self.variables[index[0]]['@']
        for i in range(2, len(index) - 2):
            ptr = ptr[index[i]]
        ptr[index[-1]] = value

    def select(self, steps: List[Union[IndexExpr, RangeExpr]]):
        # steps.0 must be index because there is no structure change
        assert isinstance(steps[0], IndexExpr)
        if steps[0].val == "@":
            # select metadata
            value = self.metadata
            for i in range(1, len(steps)):
                assert isinstance(steps[i], IndexExpr)
                value = value[steps[i].val]
            return value

        # select variable
        variable = self.variables[steps[0].val]
        assert isinstance(steps[1], IndexExpr)
        if steps[1].val == '@':
            # select metadata of variable
            value = variable['@']
            for i in range(2, len(steps)):
                assert isinstance(steps[i], IndexExpr)
                value = value[steps[i].val]
            return value

        # select data of variables
        assert steps[1].val == 'data'
        if not isinstance(variable['data'], np.ndarray):
            # TODO: we should lazy load numpy array
            variable['data'] = np.asarray(variable['data'])

        # slicing through numpy array is a piece of cake
        value = variable['data']
        for i in range(2, len(steps)):
            if isinstance(steps[i], IndexExpr):
                value = value[steps[i].val]
            else:
                value = value[steps[i].start:steps[i].end:steps[i].step]
        return value


from dataclasses import dataclass
from typing import List, Dict, Iterable, Any, Tuple, TYPE_CHECKING, Union, Optional

import numpy as np

from drepr.models import Alignment, defaultdict, RangeAlignment, ClassNode

from drepr.outputs.array_backend.array_attr import Attribute, ArrayAttr, ScalarAttr
from drepr.outputs.array_backend.array_predicate import ArrayObjectPredicate, ArrayDataPredicate, ArrayPredicate
from drepr.outputs.array_backend.subset_array_class import SubsetArrayClass
from drepr.outputs.base_output_class import BaseOutputClass
from drepr.outputs.base_output_predicate import BaseOutputPredicate
from drepr.outputs.base_output_sm import FConditionOp, FCondition
from drepr.outputs.record_id import RecordID, URIRecordID, BlankRecordID

if TYPE_CHECKING:
    from drepr.outputs.array_backend.array_backend import ArrayBackend
from drepr.outputs.array_backend.array_record import ArrayRecord
from drepr.outputs.array_backend.index_map_func import O2ORange0Func, X2OFunc, IdentityFunc, X2MFunc


class ArrayClass(BaseOutputClass):
    def __init__(self, backend: 'ArrayBackend', node_id: str):
        self.backend = backend
        # the original semantic model
        self.sm = backend.sm
        # @inherit: id of the node in the semantic model
        self.id = node_id
        # uri of the class
        self.uri = backend.sm.nodes[node_id].label
        # id of the attribute which is selected as the subject of the class
        self.pk_attr_id: str
        # the attribute which is selected as the subject of the class
        self.pk_attr: Attribute
        # if of the attribute which contains URIs of the class
        self.uri_attr_id: Optional[str] = None
        # the wrapper of the attribute which contains URIs of the class.
        self.uri_attr: PolymorphismAttribute
        # a mapping from predicate uris to the list of predicates of this class
        self.predicates: Dict[str, Union[ArrayDataPredicate, ArrayObjectPredicate]] = {}
        # the list of wrapped attributes which contains data of predicates of this class
        self.attrs: List[PolymorphismAttribute] = []
        # a mapping from predicate uris to the index of the wrapped attributes
        self.pred2attrs: Dict[str, List[int]] = defaultdict(list)

    def _init_schema(self):
        uri2edges = defaultdict(list)
        uri2type = {}
        for e in self.sm.iter_outgoing_edges(self.id):
            if isinstance(self.sm.nodes[e.target_id], ClassNode):
                if e.label in uri2type:
                    assert uri2type[e.label] == "object", "Violate the assumption"
                else:
                    uri2type[e.label] = "object"
                uri2edges[e.label].append(e)
            else:
                if e.label == 'drepr:uri':
                    self.uri_attr_id = e.target_id
                else:
                    if e.label in uri2type:
                        assert uri2type[e.label] == "data", "Violate the assumption"
                    else:
                        uri2type[e.label] = "data"
                    uri2edges[e.label].append(e)

            if e.is_subject:
                # this requires us to analyze the d-repr output first.
                self.pk_attr_id = e.target_id

        for uri, edges in uri2edges.items():
            if uri2type[uri] == "object":
                self.predicates[uri] = ArrayObjectPredicate(self.backend, edges)
            else:
                self.predicates[uri] = ArrayDataPredicate(self.backend, edges)

    def _init_data(self):
        self.pk_attr = self.backend.attrs[self.pk_attr_id]

        if self.uri_attr_id is None:
            # blank class
            self.uri_attr = PolymorphismAttribute(self.pk_attr, IdentityFunc(), True, True, self.id)
        else:
            self.uri_attr = PolymorphismAttribute(self.backend.attrs[self.uri_attr_id],
                                                  self._get_imfunc(
                                                      self.backend.alignments[self.pk_attr.id, self.uri_attr_id]),
                                                  True, False, self.id)

        # contains both values of data property and object property.
        self.attrs = []
        self.pred2attrs: Dict[str, List[int]] = defaultdict(list)
        for p in self.predicates.values():
            if isinstance(p, ArrayDataPredicate):
                for i, e in enumerate(p.edges):
                    self.pred2attrs[p.uri].append(len(self.attrs))

                    if self.pk_attr.id == p.attr(i).id:
                        imfunc = IdentityFunc()
                    else:
                        imfunc = self._get_imfunc(self.backend.alignments[self.pk_attr.id, p.attr(i).id])
                    self.attrs.append(PolymorphismAttribute(
                        p.attr(i), imfunc, True
                    ))
            else:
                p._init(self.backend)
                for i, e in enumerate(p.edges):
                    self.pred2attrs[p.uri].append(len(self.attrs))
                    imfunc = self._get_imfunc(self.backend.alignments[self.pk_attr.id, p.attr(i).id])
                    self.attrs.append(PolymorphismAttribute(
                        p.attr(i), imfunc,
                        imfunc.is_x2o, p.targets[i].is_blank(), p.targets[i].id,
                    ))

    def is_blank(self) -> bool:
        return self.uri_attr_id is None

    def iter_records(self):
        """
        Iterate through all records of this class
        """
        if isinstance(self.pk_attr, ArrayAttr):
            if self.pk_attr.nodata is not None:
                for index in np.ndindex(*self.pk_attr.values_shp):
                    if self.uri_attr.get_sval(index) is None:
                        continue
                    yield ArrayRecord(index, self)
            else:
                for index in np.ndindex(*self.pk_attr.values_shp):
                    yield ArrayRecord(index, self)
        else:
            yield ArrayRecord((0,), self)

    def get_record_by_id(self, rid: RecordID):
        """
        Get a record of the class by id
        """
        return ArrayRecord(rid.index, self)

    def p(self, predicate_uri: str) -> Optional[BaseOutputPredicate]:
        return self.predicates.get(predicate_uri, None)

    def filter(self, conditions: List[FCondition]) -> 'ArrayClass':
        if type(conditions) is list:
            if len(conditions) > 1:
                raise NotImplementedError()
            condition = conditions[0]
        else:
            condition = conditions

        if condition.op == FConditionOp.eq:
            matched_attrs = [self.attrs[a] for a in self.pred2attrs[condition.predicate_uri]]
            if sum((a.attr.size for a in matched_attrs)) == 1:
                val = [a.get_first_value() for a in matched_attrs][0]
                if val == condition.val:
                    return self
                return SubsetArrayClass(self, [])

        raise NotImplementedError()

    def group_by(self, predicate: Union[str, ArrayPredicate]) -> Iterable[Tuple[Any, 'ArrayClass']]:
        """Group by predicate (can be string or predicate id)"""
        if isinstance(predicate, str):
            # perform group by the predicate uri
            if len(self.pred2attrs[predicate]) > 1:
                raise Exception("Key of group_by operator cannot be a list")
            # p = self.p(predicate)[0]
            o = self.attrs[self.pred2attrs[predicate][0]]
        else:
            o = predicate.attr

        if o.size == 1:
            yield o.get_first_value(), self
        else:
            groups = defaultdict(set)
            for id in np.ndindex(self.pk_attr.shape):
                for val in o.get_mval(id):
                    groups[val].add(id)
            for val, ids in groups.items():
                yield val, SubsetArrayClass(self, ids)

    def _get_imfunc(self, alignments: List[Alignment]) -> Union[X2OFunc, X2MFunc]:
        # constraint of the array-backend class, we have it until we figure out how to
        # handle chained join or value join efficiency. This constraint is supposed to
        # be always satisfied since it has been checked before writing data to array-backend
        # format
        assert len(alignments) == 1 and isinstance(alignments[0], RangeAlignment)
        # have to remove fixed dimension (dimension of just one element)
        # e.g.: $.[:].lat.data[0][:][:] => 3 dims: (0, 4, 5) => remove to be (0, 1, 2)
        # the shift will  be: (0, 3, 3)
        # instead we can rely on the step2dim
        # TODO: fix me! handle the attribute id is not ideal.
        if alignments[0].source not in self.backend.attrs:
            attr = self.backend.attrs[f'dnode:{alignments[0].source}']
        else:
            attr = self.backend.attrs[alignments[0].source]

        if isinstance(attr, ScalarAttr):
            assert len(alignments[0].aligned_steps) == 0
            step2dim = []
        else:
            step2dim = attr.step2dim
        target2source = [step2dim[s.source_idx] for s in sorted(alignments[0].aligned_steps, key=lambda s: s.target_idx)]
        return O2ORange0Func(target2source)


@dataclass
class PolymorphismAttribute:
    """
    Represent attribute of one of possible three types: data property, blank object property, and uri object property
    """
    attr: Attribute
    im_func: Union[X2OFunc, X2MFunc]
    is_x2o_func: bool
    is_blank: bool = False
    class_id: str = None

    @property
    def size(self):
        return self.attr.size

    def get_first_value(self):
        zero_idx = next(np.ndindex(*self.attr.shape))
        if self.class_id is not None:
            if self.is_blank:
                return BlankRecordID(zero_idx, self.class_id)
            else:
                return RecordID(self.attr.get_value(zero_idx), zero_idx, self.class_id)
        return self.attr.get_value(zero_idx)

    def get_sval(self, index: Tuple[int, ...]):
        if self.class_id is not None:
            if self.is_blank:
                return BlankRecordID(
                    self.im_func(index) if self.is_x2o_func else next(self.im_func(index)),
                    self.class_id)
            else:
                if self.is_x2o_func:
                    idx = self.im_func(index)
                    val = self.attr.get_value(idx)
                    if self.attr.nodata is not None and self.attr.nodata.value == val:
                        return None
                else:
                    if self.attr.nodata is not None:
                        for idx in next(self.im_func(index)):
                            val = self.attr.get_value(idx)
                            if val != self.attr.nodata.value:
                                break
                        else:
                            return None
                    else:
                        idx = next(self.im_func(index))
                        val = self.attr.get_value(idx)

                return URIRecordID(val, idx, self.class_id)
        else:
            if self.is_x2o_func:
                val = self.attr.get_value(self.im_func(index))
                if self.attr.nodata is not None and val == self.attr.nodata.value:
                    return None
                return val

            if self.attr.nodata is not None:
                for val in self.im_func(index):
                    if val != self.attr.nodata.value:
                        return val
            else:
                return self.attr.get_value(next(self.im_func(index)))

    def get_mval(self, index: Tuple[int, ...]):
        if self.class_id is not None:
            if self.is_blank:
                if self.is_x2o_func:
                    return [BlankRecordID(self.im_func(index), self.class_id)]
                else:
                    return [BlankRecordID(x, self.class_id) for x in self.im_func(index)]
            if self.is_x2o_func:
                idx = self.im_func(index)
                val = self.attr.get_value(idx)
                if self.attr.nodata is not None and val == self.attr.nodata.value:
                    return []
                return [URIRecordID(val, idx, self.class_id)]

            if self.attr.nodata is not None:
                lst = []
                for x in self.im_func(index):
                    val = self.attr.get_value(x)
                    if val == self.attr.nodata.value:
                        continue
                    lst.append(URIRecordID(val, x, self.class_id))
                return lst
            else:
                return [URIRecordID(self.attr.get_value(x), x, self.class_id) for x in self.im_func(index)]
        else:
            if self.is_x2o_func:
                val = self.attr.get_value(self.im_func(index))
                if self.attr.nodata is not None and val == self.attr.nodata.value:
                    return []
                return [val]
            return [val for val in (self.attr.get_value(x) for x in self.im_func(index)) if val is not None]

    def set_value(self, index: Tuple[int, ...], val: Any):
        if self.is_blank:
            raise Exception("Cannot assign value to blank instances")
        if self.is_x2o_func:
            self.attr.set_value(self.im_func(index), val)
        else:
            # expect value to be an iterator
            for i, x in enumerate(self.im_func(index)):
                self.attr.set_value(x, next(val))



import re
from typing import List

from drepr.models.align import AlignedStep, RangeAlignment, ValueAlignment, AlignmentType, Alignment
from drepr.utils.validator import Validator, InputError


class AlignParser:
    """
    Alignments are defined as a list

    ```
    - type: <alignment_type>
      # other properties
    ```

    If <alignment_type> is `dimension`, we have two schemas:
        1. align by one dimension

        ```
        - type: dimension
          value: <source_attr_id>:<step_index> <-> <target_attr_id>:<step_index>
        ```

        2. align by multiple dimensions

        ```
        - type: dimension
          source: <attr_id>
          target: <attr_id>
          aligned_dims:
            - { source: <step_index>, target: <step_index> }
            - { source: <step_index>, target: <step_index> }
            ...
        ```
    
    If <alignment_type> is `value`, then the schema is
        ```
        - type: value
          source: <attr_id>
          target: <attr_id>
        ```
    """
    ALIGNMENT_TYPES_VALUES = {x.value if x != AlignmentType.range else "dimension" for x in AlignmentType}
    REG_DALIGN = re.compile(r"^(.+):(\d+) *<-> *(.+):(\d+)$")
    DALIGN_KEYS = {"type", "source", "target", "aligned_dims"}
    DALIGN_ALIGNED_DIMS_KEYS = {"source", "target"}
    VALIGN_KEYS = {"type", "source", "target"}

    @classmethod
    def parse(cls, alignments: List[dict]) -> List[Alignment]:
        Validator.must_be_list(alignments, "Parsing alignments")
        aligns = []

        for i, align in enumerate(alignments):
            trace = f"Parsing an alignment at index {i}"
            Validator.must_be_dict(align, trace)
            Validator.must_have(align, "type", trace)

            align_type = AlignmentType(align["type"] if align["type"] != "dimension" else "range")
            if align_type == AlignmentType.range:
                aligns.append(cls.parse_range_align(align, trace))
            elif align_type == AlignmentType.value:
                aligns.append(cls.parse_value_align(align, trace))
            else:
                raise NotImplemented(
                    f"{trace}\nERROR: not implement parser for alignment type: {align_type}")

        return aligns

    @classmethod
    def parse_range_align(cls, conf: dict, parse_trace: str) -> RangeAlignment:
        if "value" in conf:
            Validator.must_be_str(conf["value"],
                                  f"{parse_trace}\nParsing property `value` of the alignment")
            m = cls.REG_DALIGN.match(conf["value"])
            if m is None:
                raise InputError(
                    f"{parse_trace}\nERROR: invalid `value` property, its format is invalid")

            return RangeAlignment(m.group(1), m.group(3),
                                  [AlignedStep(int(m.group(2)), int(m.group(4)))])

        Validator.must_be_subset(cls.DALIGN_KEYS, conf.keys(), "properties of alignment",
                                 parse_trace)
        aligned_dims = []
        for i, adim in enumerate(conf['aligned_dims']):
            Validator.must_be_subset(cls.DALIGN_ALIGNED_DIMS_KEYS, adim.keys(),
                                     "properties of aligned_dims", parse_trace)
            for key in cls.DALIGN_ALIGNED_DIMS_KEYS:
                Validator.must_be_int(
                    adim[key],
                    f"{parse_trace}\nParsing property `{key}` of aligned_dim at position {i}")
            aligned_dims.append(AlignedStep(adim['source'], adim['target']))
        return RangeAlignment(conf["source"], conf["target"], aligned_dims)

    @classmethod
    def parse_value_align(cls, conf: dict, parse_trace: str) -> ValueAlignment:
        Validator.must_be_subset(cls.VALIGN_KEYS, conf.keys(), "properties of alignment",
                                 parse_trace)
        for key in cls.VALIGN_KEYS:
            Validator.must_be_str(conf[key],
                                  f"{parse_trace}\nParsing property `{key}` of value alignment")

        return ValueAlignment(conf["source"], conf["target"])



__version__ = "2.8"   # ___PKG_VERSION___: DO NOT MODIFY the version here. Update it via version_manager.py!
__engine_version__ = "1.0.8"   # ___PKG_VERSION___: DO NOT MODIFY the version here. Update it via version_manager.py!
__engine_release_tag__ = "2.7"   # ___PKG_VERSION___: DO NOT MODIFY the version here. Update it via version_manager.py!



from dataclasses import dataclass
from typing import List, Optional

import numpy as np

from drepr.outputs.array_backend.array_attr import NoData


@dataclass
class IndexPropRange:
    # start dimension
    start: int
    # end dimension
    end: int


@dataclass
class PropDataNDArray:
    data: np.ndarray
    nodata: Optional[NoData]
    index_props_range: List[IndexPropRange]
    index_props: List[np.ndarray]



from typing import List

from drepr.utils.validator import Validator, InputError
from .path_parser import PathParser
from ..attr import Attr, Sorted, ValueType
from ..resource import Resource

class AttrParser:
    """
    Attributes have two possible schemas
    1. When an attribute has only one path
        ```
        <attr_id>: <attr_path>
        # .. other attributes ..
        ```
    2.
        ```
        <attribute_id>:
            [resource_id]: <resource_id> (default is "default")
            path: <attr_path>
            [unique]: true|false (default is false)
            [sorted]: none|ascending|descending (default is none)
            [value_type]: unspecified|int|float|str|list[int]|list[str]|list[float] (default is unspecified)
            [missing_values]: [<value0>, <value1>, ...]
        ```
    """
    SORTED_VALUES = {x.value for x in Sorted}
    VALUE_TYPE_VALUES = {x.value for x in ValueType}

    def __init__(self, path_parser: PathParser):
        self.path_parser = path_parser

    def parse(self, default_resource_id: str, resources: List[Resource], attrs: dict) -> List[Attr]:
        Validator.must_be_dict(attrs, "Parsing attributes")
        result = []
        for attr_id, attr_conf in attrs.items():
            trace = f"Parsing attribute: {attr_id}"

            if isinstance(attr_conf, (str, list)):
                attr_path = self.path_parser.parse(
                    self.path_parser.get_resource(resources, default_resource_id, trace),
                    attr_conf, trace)
                attr = Attr(attr_id, default_resource_id, attr_path, [])
            elif isinstance(attr_conf, dict):
                attr = self.parse_schema2(default_resource_id, resources, attr_id, attr_conf, trace)
            else:
                raise InputError(
                    f"{trace}\nERROR: The configuration of an attribute can either be string, list, "
                    f"or dictionary. Get {type(attr_conf)} instead")
            result.append(attr)

        return result

    def parse_schema2(self, default_resource_id: str, resources: List[Resource], attr_id: str, attr_conf: dict, parse_trace: str) -> Attr:
        resource_id = attr_conf.get("resource_id", default_resource_id)
        Validator.must_be_str(resource_id, f"{parse_trace}\nParsing `resource_id`")

        Validator.must_have(attr_conf, "path", parse_trace)
        path = self.path_parser.parse(
            self.path_parser.get_resource(resources, resource_id, parse_trace),
            attr_conf['path'], f"{parse_trace}\nParsing path of the attribute")

        if "unique" in attr_conf and not isinstance(attr_conf['unique'], bool):
            raise InputError(f"{parse_trace}\nERROR: invalid value of the `unique` attribute. "
                             f"Expected a boolean value. Get: {attr_conf['unique']}")
        unique = attr_conf.get('unique', False)

        if 'sorted' in attr_conf:
            Validator.must_in(attr_conf['sorted'], self.SORTED_VALUES,
                              f"{parse_trace}\nParsing `sorted` of the attribute")
        sorted = Sorted(attr_conf.get('sorted', Sorted.Null.value))

        if 'value_type' in attr_conf:
            Validator.must_in(attr_conf['value_type'], self.VALUE_TYPE_VALUES,
                              f"{parse_trace}\nParsing `value_type` of the attribute")
        value_type = ValueType(attr_conf.get('value_type', ValueType.Unspecified.value))

        if 'missing_values' in attr_conf:
            trace = f"{parse_trace}\nParsing missing_values of the attribute"
            Validator.must_be_list(attr_conf['missing_values'], trace)
            for val in attr_conf['missing_values']:
                if not isinstance(val, (str, int, float)):
                    raise InputError(
                        f"{trace}\nERROR: invalid value. Expected either one of string, "
                        f"integer, or float. Get f{type(val)} instead")

        missing_values = attr_conf.get('missing_values', [])
        return Attr(attr_id, resource_id, path, missing_values, unique, sorted, value_type)



from typing import List, Optional, TYPE_CHECKING, Union

import numpy as np

from drepr.models import RangeAlignment, Edge
from drepr.outputs.array_backend.array_attr import Attribute
from drepr.outputs.array_backend.lst_array_class import LstArrayClass
from drepr.outputs.prop_data_ndarray import PropDataNDArray, IndexPropRange

if TYPE_CHECKING:
    from drepr.outputs.array_backend.array_backend import ArrayBackend
from drepr.outputs.base_output_predicate import BaseOutputPredicate


class ArrayDataPredicate(BaseOutputPredicate):
    def __init__(self, backend: 'ArrayBackend', edges: List[Edge]):
        self.backend = backend
        self.edges = edges
        self.uri = edges[0].label

    def as_ndarray(self, index_predicates: List[Union['ArrayDataPredicate', 'ArrayObjectPredicate']]) -> PropDataNDArray:
        """
        Get predicate data (identified by `pred_id` as a high-dimensional array). The original data may already be in high-dimension
        array or may be not, but the returned value must be a high-dimensional array.

        The supplied `index_edges` are list of edges that will occupied first dimensions. If an edge in index_edges are
        high-dimensional array as well, then its value will be flatten.

        There must be an alignment between the edge_id and other index edges (the alignment represent the join). The alignment
        must be dimension alignment for now (then we don't need to do a join but only swapping and arranging dimension). In case the
        alignment are chained, then we have to join, and create new table?
        """
        if len(self.edges) > 1 or any(len(p.edges) > 1 for p in index_predicates):
            raise Exception("Cannot convert values of this predicate to an ndarray indexed by other predicates "
                            "because values for one entry may be greater than one")

        attr = self.attr(0)
        # these index columns may also be in high-dimensional array
        index_attrs: List[Attribute] = [p.attr(0) for p in index_predicates]

        """
        1. The algorithm works by first retrieve original nd-array of the column called C.
        2. We get alignments between the column and other index columns; and make sure that the alignments are all 
        dimension alignments. If there are some alignments that are not dimension alignments, we fall back to doing join
        between multiple tables, this would be slow!.
        3. For each index column: ICi:
            1. We retrieve the aligned dimensions between the column C and the index column ICi, now if there is any 
            dimension in ICi that is not aligned (*-to-many), we have to replicate the value C to match that. 
            2. Then, we mark which dimension of C is linked to dimension of ICi.
        4. After the above step, we retrieve a nested list, each dimension in C is annotated with a bunch of aligned
           dimension of ICi. We want to swap these dimensions so that the dimensions are in order of the indexed ICi.
           However, we may run into the case where there may be overlapping between dimensions of ICi. For example:
            1 1 1 1
            |-|-|-|-|
                2 2 2 
            Then, the result dimension would be:
            1 1 1
            |-|-|-|
              2 2 2

            If we cannot make dimension of index columns continuous (e.g., mixed, or one dim have 3 index_cols), then
            we have to throw error because there is no way to handle that. It cannot be continuous when at one dimension,
            the list of indexed edges are not continuous e.g, at dim 2, it is indexed by both col 1 and 3 (missing col 2)

            In that case, the users should use other functions to handle the situation yourself.
        """
        # 1st retrieve the data
        data = attr.get_data()
        data_dims = [[] for _ in range(len(data.shape))]

        alignments: List[List[RangeAlignment]] = [self.backend.alignments[attr.id, iattr.id] for iattr in
                                                  index_attrs]
        for aligns, index_col, index_col_idx in zip(alignments, index_attrs, range(len(index_attrs))):
            # source is always col_id
            assert len(aligns) == 1
            align = aligns[0]
            col_aligned_steps = {step.source_idx for step in align.aligned_steps}
            index_col_aligned_steps = {step.target_idx for step in align.aligned_steps}

            # 3.1 find any dimension in ICi that is not aligned, but also must select more than one element (*-to-many)
            index_col_unbounded_dims = [
                dim
                for step_idx, dim in enumerate(index_col.step2dim)
                if step_idx not in index_col_aligned_steps and dim is not None
            ]
            if len(index_col_unbounded_dims) > 0:
                # calculate the number of elements k we need to repeated
                k = np.prod([index_col.shape[i] for i in index_col_unbounded_dims])
                data = data.repeat(k, axis=-1)
                # perform step 3.2 here
                data_dims.append([index_col_idx])

            # 3.2 mark dimension of data, saying which dimension is linked to which dimension of ICi
            for col_step_idx in col_aligned_steps:
                col_dim = attr.step2dim[col_step_idx]
                data_dims[col_dim].append(index_col_idx)

        # now is the swapping dimensions part, although numpy may just return different view,
        # we don't want to swap a lot because if we can swap freely, why numpy transpose cannot guarantee for returning
        # a view?
        # detect when we cannot swap
        for icis in data_dims:
            if len(icis) > 0 and len(icis) != (icis[-1] - icis[0] + 1):
                raise Exception("Cannot find a way to satisfied the query condition")

        # find the new order of dimensions by sorting
        new_data_dims = []
        new_axies = []
        axis = list(range(len(data_dims)))
        for i in range(len(data_dims)):
            min_j = i
            for j in range(i, len(data_dims)):
                # min(data_dims[*]) = data_dims[*][0]
                if len(data_dims[j]) > 0 and data_dims[j][0] < data_dims[i][0]:
                    min_j = j
            if min_j != i:
                # we have to swap
                data_dims[min_j], data_dims[i] = data_dims[i], data_dims[min_j]
                axis[min_j], axis[i] = axis[i], axis[min_j]
            new_data_dims.append(data_dims[i])
            new_axies.append(axis[i])

        # finally, return the range of each index columns from the new_data_dims to return
        index_attr_positions = [IndexPropRange(len(index_attrs) * 2, 0) for _ in range(len(index_attrs))]
        for i, icis in enumerate(new_data_dims):
            for j in icis:
                index_attr_positions[j].start = min(i, index_attr_positions[j].start)
                index_attr_positions[j].end = max(i + 1, index_attr_positions[j].end)
        data = np.transpose(data, new_axies)
        return PropDataNDArray(data, attr.nodata, index_attr_positions, [a.get_data() for a in index_attrs])

    def o(self) -> Optional['LstArrayClass']:
        return None

    def attr(self, idx: int):
        return self.backend.attrs[self.edges[idx].target_id]


class ArrayObjectPredicate(BaseOutputPredicate):
    def __init__(self, backend: 'ArrayBackend', edges: List[Edge]):
        self.uri = edges[0].label
        self.edges = edges
        self.targets = LstArrayClass([backend.cid(e.target_id) for e in edges])

    def _init(self, backend):
        self.attrs = [
            backend.attrs[target.pk_attr_id] if target.is_blank()
            else backend.attrs[target.uri_attr_id]
            for target in self.targets
        ]

    def as_ndarray(self, indexed_predicates: List['ArrayPredicate']) -> PropDataNDArray:
        raise NotImplementedError()

    def o(self) -> Optional['LstArrayClass']:
        return self.targets

    def attr(self, idx: int):
        return self.attrs[idx]


ArrayPredicate = Union[ArrayDataPredicate, ArrayObjectPredicate]


import os
import shutil
import sys
import urllib.request
from setuptools import setup, find_packages
from setuptools.command.install import install

wdir = os.path.dirname(os.path.abspath(__file__))

# load versions
exec(open("drepr/version.py").read())

if sys.argv[1] != "sdist":
    # download correct engine version, only run if this is not the command that build the dist
    target_triple = {
        "linux": "x86_64-unknown-linux-gnu",
        "darwin": "x86_64-apple-darwin",
        "win32": "x86_64-pc-windows-msvc"
    }[sys.platform]
    pylib_ext = {"linux": "so", "darwin": "so", "win32": "pyd"}[sys.platform]
    engine_url = f"https://github.com/usc-isi-i2/d-repr/releases/download/{__engine_release_tag__}/drepr_engine.{target_triple}.{__engine_version__}.{pylib_ext}"
    is_downloaded_test_file = f"{wdir}/download-executed-gn13v9a5.txt"

    if os.path.exists(is_downloaded_test_file):
        print(">>> pre-built engine is already exist! Skip it")
    else:
        print(">>> download pre-built engine at", engine_url)
        urllib.request.urlretrieve(engine_url, f"{wdir}/drepr/drepr_engine.{pylib_ext}")
        with open(is_downloaded_test_file, "w") as f:
            f.write("downloaded")

with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), "README.md"), "r", encoding="utf-8") as f:
    long_description = f.read()

setup(name="drepr",
      version=__version__,
      packages=find_packages(),
      description="Data Representation Language for Reading Heterogeneous Datasets",
      long_description=long_description,
      long_description_content_type='text/markdown',
      author="Binh Vu",
      author_email="binh@toan2.com",
      url="https://github.com/usc-isi-i2/d-repr",
      python_requires='>3.6',
      license="MIT",
      install_requires=['ujson', 'ruamel.yaml>=0.15.0',
                        'dataclasses;python_version<"3.7"', 'xmltodict', 'netcdf4', 'fiona',
                        'pillow'],
      package_data={'': ['*.so', '*.pyd']})
